@@ -15,6 +15,8 @@
  */
 
 #include "apxwin.h"
+#include "handles.h"
+#include "javajni.h"
 #include "private.h"
 
 #include <jni.h>
@@ -61,7 +63,6 @@ static DYNLOAD_FPTR_DECLARE(SetDllDirectoryW) = NULL;
 #define JVM_EXCEPTION_CHECK(jvm) \
     ((*((jvm)->lpEnv))->ExceptionCheck((jvm)->lpEnv) != JNI_OK)
 
-#ifdef _DEBUG
 #define JVM_EXCEPTION_CLEAR(jvm) \
     APXMACRO_BEGIN                                              \
     if ((jvm)->lpEnv) {                                         \
@@ -70,15 +71,6 @@ static DYNLOAD_FPTR_DECLARE(SetDllDirectoryW) = NULL;
             (*((jvm)->lpEnv))->ExceptionClear((jvm)->lpEnv);    \
         }                                                       \
     } APXMACRO_END
-#else
-#define JVM_EXCEPTION_CLEAR(jvm) \
-    APXMACRO_BEGIN                                              \
-    if ((jvm)->lpEnv) {                                         \
-        if ((*((jvm)->lpEnv))->ExceptionCheck((jvm)->lpEnv)) {  \
-            (*((jvm)->lpEnv))->ExceptionClear((jvm)->lpEnv);    \
-        }                                                       \
-    } APXMACRO_END
-#endif
 
 #define JNI_LOCAL_UNREF(obj) \
         (*(lpJava->lpEnv))->DeleteLocalRef(lpJava->lpEnv, obj)
@@ -119,12 +111,33 @@ typedef struct APXJAVAVM {
     HANDLE          hWorkerThread;
     DWORD           iWorkerThread;
     DWORD           dwWorkerStatus;
-
+    SIZE_T          szStackSize;
+    HANDLE          hWorkerSync;
 } APXJAVAVM, *LPAPXJAVAVM;
 
+/* This is no longer exported in jni.h
+ * However java uses it internally to get
+ * the default stack size
+ */
+typedef struct APX_JDK1_1InitArgs {
+    jint version;
+
+    char **properties;
+    jint checkSource;
+    jint nativeStackSize;
+    jint javaStackSize;
+    jint minHeapSize;
+    jint maxHeapSize;
+    jint verifyMode;
+    char *classpath;
+
+    char padding[128];
+} APX_JDK1_1InitArgs;
+
 #define JAVA_CLASSPATH      "-Djava.class.path="
 #define JAVA_CLASSPATH_W    L"-Djava.class.path="
 #define JAVA_CLASSSTRING    "java/lang/String"
+#define MSVCRT71_DLLNAME    L"\\msvcrt71.dll"
 
 static __inline BOOL __apxJvmAttach(LPAPXJAVAVM lpJava)
 {
@@ -163,6 +176,7 @@ static __inline BOOL __apxJvmDetach(LPAPXJAVAVM lpJava)
 static BOOL __apxLoadJvmDll(LPCWSTR szJvmDllPath)
 {
     UINT errMode;
+    WCHAR  jreAltPath[SIZ_PATHLEN];
     LPWSTR dllJvmPath = (LPWSTR)szJvmDllPath;
     DYNLOAD_FPTR_DECLARE(SetDllDirectoryW);
 
@@ -173,10 +187,52 @@ static BOOL __apxLoadJvmDll(LPCWSTR szJvmDllPath)
         dllJvmPath = apxGetJavaSoftRuntimeLib(NULL);
     if (!dllJvmPath)
         return FALSE;
+    if (GetFileAttributesW(dllJvmPath) == INVALID_FILE_ATTRIBUTES) {
+        /* DAEMON-184: RuntimeLib registry key is invalid.
+         * Check from Jre JavaHome directly
+         */
+        LPWSTR szJreHome = apxGetJavaSoftHome(NULL, TRUE);
+        apxLogWrite(APXLOG_MARK_DEBUG "Invalid RuntimeLib '%S'", dllJvmPath);
+        if (szJreHome) {
+            apxLogWrite(APXLOG_MARK_DEBUG "Using Jre JavaHome '%S'", szJreHome);
+            lstrlcpyW(jreAltPath, SIZ_PATHLEN, szJreHome);
+            lstrlcatW(jreAltPath, SIZ_PATHLEN, L"\\bin\\server\\jvm.dll");
+            dllJvmPath = jreAltPath;
+        }
+    }
     /* Suppress the not found system popup message */
     errMode = SetErrorMode(SEM_FAILCRITICALERRORS);
 
+    apxLogWrite(APXLOG_MARK_DEBUG "loading jvm '%S'", dllJvmPath);
     _st_sys_jvmDllHandle = LoadLibraryExW(dllJvmPath, NULL, 0);
+    if (IS_INVALID_HANDLE(_st_sys_jvmDllHandle) &&
+        GetFileAttributesW(dllJvmPath) != INVALID_FILE_ATTRIBUTES) {
+        /* There is a file but cannot be loaded.
+         * Try to load the MSVCRTxx.dll before JVM.dll
+         */
+        WCHAR  jreBinPath[SIZ_PATHLEN];
+        WCHAR  crtBinPath[SIZ_PATHLEN];
+        DWORD  i, l = 0;
+
+        lstrlcpyW(jreBinPath, SIZ_PATHLEN, dllJvmPath);
+        for (i = lstrlenW(jreBinPath); i > 0, l < 2; i--) {
+            if (jreBinPath[i] == L'\\' || jreBinPath[i] == L'/') {
+                jreBinPath[i] = L'\0';
+                lstrlcpyW(crtBinPath, SIZ_PATHLEN, jreBinPath);
+                lstrlcatW(crtBinPath, SIZ_PATHLEN, MSVCRT71_DLLNAME);
+                if (GetFileAttributesW(crtBinPath) != INVALID_FILE_ATTRIBUTES) {
+                    if (LoadLibraryW(crtBinPath)) {
+                        /* Found MSVCRTxx.dll
+                         */
+                        apxLogWrite(APXLOG_MARK_DEBUG "preloaded '%S'",
+                                    crtBinPath);
+                        break;
+                    }
+                }
+                l++;
+            }
+        }
+    }
     /* This shuldn't happen, but try to search in %PATH% */
     if (IS_INVALID_HANDLE(_st_sys_jvmDllHandle))
         _st_sys_jvmDllHandle = LoadLibraryExW(dllJvmPath, NULL,
@@ -192,6 +248,8 @@ static BOOL __apxLoadJvmDll(LPCWSTR szJvmDllPath)
             if (jreBinPath[i] == L'\\' || jreBinPath[i] == L'/') {
                 jreBinPath[i] = L'\0';
                 DYNLOAD_CALL(SetDllDirectoryW)(jreBinPath);
+                apxLogWrite(APXLOG_MARK_DEBUG "Setting DLL search path to '%S'",
+                            jreBinPath);
                 l++;
             }
         }
@@ -267,9 +325,12 @@ apxCreateJava(APXHANDLE hPool, LPCWSTR szJvmDllPath)
     LPAPXJAVAVM  lpJava;
     jsize        iVmCount;
     JavaVM       *lpJvm = NULL;
+    struct       APX_JDK1_1InitArgs jArgs1_1;
 
     if (!__apxLoadJvmDll(szJvmDllPath))
         return NULL;
+
+
     /*
      */
     if (DYNLOAD_FPTR(JNI_GetCreatedJavaVMs)(&lpJvm, 1, &iVmCount) != JNI_OK) {
@@ -287,6 +348,16 @@ apxCreateJava(APXHANDLE hPool, LPCWSTR szJvmDllPath)
     lpJava = APXHANDLE_DATA(hJava);
     lpJava->lpJvm = lpJvm;
     lpJava->iVmCount = iVmCount;
+
+    /* Guess the stack size
+     */
+    AplZeroMemory(&jArgs1_1, sizeof(jArgs1_1));
+    jArgs1_1.version = JNI_VERSION_1_1;
+    DYNLOAD_FPTR(JNI_GetDefaultJavaVMInitArgs)(&jArgs1_1);
+    if (jArgs1_1.javaStackSize < 0 || jArgs1_1.javaStackSize > (2048 * 1024))
+        jArgs1_1.javaStackSize = 0;
+    lpJava->szStackSize = (SIZE_T)jArgs1_1.javaStackSize;
+
     if (!_st_sys_jvm)
         _st_sys_jvm = lpJvm;
     return hJava;
@@ -338,10 +409,7 @@ static DWORD __apxMultiSzToJvmOptions(APXHANDLE hPool,
         l = __apxGetMultiSzLengthA(lpString, &n);
     }
     n += nExtra;
-    if (IS_INVALID_HANDLE(hPool))
-        buff = apxPoolAlloc(hPool, (n + 1) * sizeof(JavaVMOption) + (l + 1));
-    else
-        buff = apxAlloc((n + 1) * sizeof(JavaVMOption) + (l + 1));
+    buff = apxPoolAlloc(hPool, (n + 1) * sizeof(JavaVMOption) + (l + 1));
 
     *lppArray = (JavaVMOption *)buff;
     p = (LPSTR)(buff + (n + 1) * sizeof(JavaVMOption));
@@ -610,24 +678,10 @@ apxJavaInitialize(APXHANDLE hJava, LPCSTR szClassPath,
         apxFree(szCp);
         apxFree(lpJvmOptions);
     }
-    /* Load standard classes */
-    if (rv) {
-        jclass jClazz = JNICALL_1(FindClass, JAVA_CLASSSTRING);
-        if (!jClazz) {
-            apxLogWrite(APXLOG_MARK_ERROR "FindClass "  JAVA_CLASSSTRING " failed");
-            goto cleanup;
-        }
-        lpJava->clString.jClazz = JNICALL_1(NewGlobalRef, jClazz);
-        JNI_LOCAL_UNREF(jClazz);
-
+    if (rv)
         return TRUE;
-    }
     else
         return FALSE;
-
-cleanup:
-    JVM_EXCEPTION_CLEAR(lpJava);
-    return FALSE;
 }
 
 /* ANSI version only */
@@ -733,13 +787,31 @@ apxJavaLoadMainClass(APXHANDLE hJava, LPCSTR szClassName,
     DWORD       nArgs;
     LPAPXJAVAVM lpJava;
     jclass      jClazz;
+    LPCSTR      szSignature = "([Ljava/lang/String;)V";
 
     if (hJava->dwType != APXHANDLE_TYPE_JVM)
         return FALSE;
     lpJava = APXHANDLE_DATA(hJava);
-    if (!__apxJvmAttach(lpJava))
+    if (!lpJava)
         return FALSE;
+    if (IS_EMPTY_STRING(szMethodName))
+        szMethodName = "main";
+    if (lstrcmpA(szClassName, "java/lang/System") == 0) {
+        /* Usable only for exit method, so force */
+        szSignature  = "(I)V";
+        szMethodName = "exit";
+    }
+    lstrlcpyA(lpJava->clWorker.sClazz, 1024, szClassName);
+    lstrlcpyA(lpJava->clWorker.sMethod, 512, szMethodName);
 
+    jClazz = JNICALL_1(FindClass, JAVA_CLASSSTRING);
+    if (!jClazz) {
+        JVM_EXCEPTION_CLEAR(lpJava);
+        apxLogWrite(APXLOG_MARK_ERROR "FindClass "  JAVA_CLASSSTRING " failed");
+        return FALSE;
+    }
+    lpJava->clString.jClazz = JNICALL_1(NewGlobalRef, jClazz);
+    JNI_LOCAL_UNREF(jClazz);
     /* Find the class */
     jClazz  = JNICALL_1(FindClass, szClassName);
     if (!jClazz) {
@@ -751,31 +823,29 @@ apxJavaLoadMainClass(APXHANDLE hJava, LPCSTR szClassName,
     lpJava->clWorker.jClazz  = JNICALL_1(NewGlobalRef, jClazz);
     JNI_LOCAL_UNREF(jClazz);
 
-    if (!szMethodName)
-        szMethodName = "main";
-    lstrlcpyA(lpJava->clWorker.sClazz, 1024, szClassName);
-    lstrlcpyA(lpJava->clWorker.sMethod, 512, szMethodName);
     lpJava->clWorker.jMethod = JNICALL_3(GetStaticMethodID,
                                          lpJava->clWorker.jClazz,
-                                         szMethodName, "([Ljava/lang/String;)V");
+                                         szMethodName, szSignature);
     if (!lpJava->clWorker.jMethod) {
         JVM_EXCEPTION_CLEAR(lpJava);
         apxLogWrite(APXLOG_MARK_ERROR "Method 'static void %s(String[])' not found in Class %s",
                 szMethodName, szClassName);
         return FALSE;
     }
-    nArgs = apxMultiSzToArrayW(hJava->hPool, lpArguments, &lpArgs);
-    lpJava->clWorker.jArgs = JNICALL_3(NewObjectArray, nArgs,
-                                       lpJava->clString.jClazz, NULL);
-    if (nArgs) {
-        DWORD i;
-        for (i = 0; i < nArgs; i++) {
-            jstring arg = JNICALL_2(NewString, lpArgs[i], lstrlenW(lpArgs[i]));
-            JNICALL_3(SetObjectArrayElement, lpJava->clWorker.jArgs, i, arg);
-            apxLogWrite(APXLOG_MARK_DEBUG "argv[%d] = %S", i, lpArgs[i]);
+    if (lstrcmpA(szClassName, "java/lang/System")) {
+        nArgs = apxMultiSzToArrayW(hJava->hPool, lpArguments, &lpArgs);
+        lpJava->clWorker.jArgs = JNICALL_3(NewObjectArray, nArgs,
+                                           lpJava->clString.jClazz, NULL);
+        if (nArgs) {
+            DWORD i;
+            for (i = 0; i < nArgs; i++) {
+                jstring arg = JNICALL_2(NewString, lpArgs[i], lstrlenW(lpArgs[i]));
+                JNICALL_3(SetObjectArrayElement, lpJava->clWorker.jArgs, i, arg);
+                apxLogWrite(APXLOG_MARK_DEBUG "argv[%d] = %S", i, lpArgs[i]);
+            }
         }
+        apxFree(lpArgs);
     }
-    apxFree(lpArgs);
     return TRUE;
 }
 
@@ -785,61 +855,101 @@ apxJavaLoadMainClass(APXHANDLE hJava, LPCSTR szClassName,
  */
 static DWORD WINAPI __apxJavaWorkerThread(LPVOID lpParameter)
 {
-#define WORKER_EXIT(x)  { rv = x; goto finished; }
+#define WORKER_EXIT(x)  do { rv = x; goto finished; } while(0)
     DWORD rv = 0;
-    LPAPXJAVAVM lpJava;
-    APXHANDLE   hJava = (APXHANDLE)lpParameter;
-    /* This shouldn't happen */
+    LPAPXJAVAVM lpJava = NULL;
+    LPAPXJAVA_THREADARGS pArgs = (LPAPXJAVA_THREADARGS)lpParameter;
+    APXHANDLE hJava;
+
+    hJava  = (APXHANDLE)pArgs->hJava;
     if (hJava->dwType != APXHANDLE_TYPE_JVM)
-        WORKER_EXIT(0);
-    lpJava = APXHANDLE_DATA(hJava);
+        WORKER_EXIT(1);
+    lpJava = APXHANDLE_DATA(pArgs->hJava);
+    if (!lpJava)
+        WORKER_EXIT(1);
+    if (!apxJavaInitialize(pArgs->hJava,
+                           pArgs->szClassPath,
+                           pArgs->lpOptions,
+                           pArgs->dwMs, pArgs->dwMx, pArgs->dwSs,
+                           pArgs->bJniVfprintf)) {
+        WORKER_EXIT(2);
+    }
+    if (pArgs->szLibraryPath && *pArgs->szLibraryPath) {
+        DYNLOAD_FPTR_ADDRESS(SetDllDirectoryW, KERNEL32);
+        DYNLOAD_CALL(SetDllDirectoryW)(pArgs->szLibraryPath);
+        apxLogWrite(APXLOG_MARK_DEBUG "DLL search path set to '%S'",
+                    pArgs->szLibraryPath);
+    }
+    if (!apxJavaLoadMainClass(pArgs->hJava,
+                              pArgs->szClassName,
+                              pArgs->szMethodName,
+                              pArgs->lpArguments)) {
+        WORKER_EXIT(3);
+    }
+    apxJavaSetOut(pArgs->hJava, TRUE,  pArgs->szStdErrFilename);
+    apxJavaSetOut(pArgs->hJava, FALSE, pArgs->szStdOutFilename);
+
     /* Check if we have a class and a method */
     if (!lpJava->clWorker.jClazz || !lpJava->clWorker.jMethod)
-        WORKER_EXIT(2);
+        WORKER_EXIT(4);
     if (!__apxJvmAttach(lpJava))
-        WORKER_EXIT(3);
+        WORKER_EXIT(5);
     apxLogWrite(APXLOG_MARK_DEBUG "Java Worker thread started %s:%s",
                 lpJava->clWorker.sClazz, lpJava->clWorker.sMethod);
     lpJava->dwWorkerStatus = 1;
+    SetEvent(lpJava->hWorkerSync);
     JNICALL_3(CallStaticVoidMethod,
               lpJava->clWorker.jClazz,
               lpJava->clWorker.jMethod,
               lpJava->clWorker.jArgs);
-
-    JVM_EXCEPTION_CLEAR(lpJava);
-    __apxJvmDetach(lpJava);
-    apxLogWrite(APXLOG_MARK_DEBUG "Java Worker thread %s:%s finished",
-                lpJava->clWorker.sClazz, lpJava->clWorker.sMethod);
+    if (JVM_EXCEPTION_CHECK(lpJava)) {
+        WORKER_EXIT(6);
+    }
+    else {
+        __apxJvmDetach(lpJava);
+    }
 finished:
-    lpJava->dwWorkerStatus = 0;
-    apxLogWrite(APXLOG_MARK_DEBUG "Java Worker thread finished %s:%s",
-                lpJava->clWorker.sClazz, lpJava->clWorker.sMethod);
+    if (lpJava) {
+        lpJava->dwWorkerStatus = 0;
+        apxLogWrite(APXLOG_MARK_DEBUG "Java Worker thread finished %s:%s with status=%d",
+                    lpJava->clWorker.sClazz, lpJava->clWorker.sMethod, rv);
+        SetEvent(lpJava->hWorkerSync);
+    }
     ExitThread(rv);
     /* never gets here but keep the compiler happy */
-    return 0;
+    return rv;
 }
 
-
 BOOL
-apxJavaStart(APXHANDLE hJava)
+apxJavaStart(LPAPXJAVA_THREADARGS pArgs)
 {
-
     LPAPXJAVAVM lpJava;
-
-    if (hJava->dwType != APXHANDLE_TYPE_JVM)
+    lpJava = APXHANDLE_DATA(pArgs->hJava);
+    if (!lpJava)
         return FALSE;
-    lpJava = APXHANDLE_DATA(hJava);
-
-    lpJava->hWorkerThread = CreateThread(NULL, 0, __apxJavaWorkerThread,
-                                         hJava, CREATE_SUSPENDED,
-                                         &lpJava->iWorkerThread);
+    lpJava->dwWorkerStatus = 0;
+    lpJava->hWorkerSync    = CreateEvent(NULL, FALSE, FALSE, NULL);
+    lpJava->hWorkerThread  = CreateThread(NULL,
+                                          lpJava->szStackSize,
+                                          __apxJavaWorkerThread,
+                                          pArgs, CREATE_SUSPENDED,
+                                          &lpJava->iWorkerThread);
     if (IS_INVALID_HANDLE(lpJava->hWorkerThread)) {
         apxLogWrite(APXLOG_MARK_SYSERR);
         return FALSE;
     }
     ResumeThread(lpJava->hWorkerThread);
-    /* Give some time to initialize the thread */
-    Sleep(1000);
+    /* Wait until the worker thread initializes */
+    WaitForSingleObject(lpJava->hWorkerSync, INFINITE);
+    if (lpJava->dwWorkerStatus == 0)
+        return FALSE;
+    if (lstrcmpA(lpJava->clWorker.sClazz, "java/lang/System")) {
+        /* Give some time to initialize the thread
+         * Unless we are calling System.exit(0).
+         * This will be hanled by _onexit hook.
+         */
+        Sleep(1000);
+    }
     return TRUE;
 }
 
